/***************************************************************************
 *   Copyright (C) 2006 by AST   *
 *   tsyplyaev@gmail.com   *
 *   ICQ: 279-533-134                          *
 *   This is a part of work done by AST.       *
 *   If you want to use it, please contact me. *
 ***************************************************************************/

#include <max.h>
#include <stdmat.h>
#include <vector>
#include <iparamm2.h>
#include <modstack.h>
#include <iskin.h>
#include "String.h"

#pragma comment(lib, "core.lib") 
#pragma comment(lib, "maxutil.lib")
#pragma comment(lib, "geom.lib")
#pragma comment(lib, "mesh.lib")
#pragma comment(lib, "comctl32.lib")

struct Weight {
	int bone;
	float weight;
	Point3 position;
	Point3 normal;
};

struct Vertex {
	Point3 position;
	Point3 normal;
	Point2 texcoord;

	int numWeights;
	Weight weights[32];

	int flag;
	int id;
};

//---------------------------------------------------------------------------
//AST ENUM PROC
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//AST ENUM PROC
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//AST ENUM PROC
//---------------------------------------------------------------------------
class AstEnumProc : public ITreeEnumProc {
public:
	AstEnumProc(Interface *i) { iface = i; };

	int callback(INode *node);
	void exportAll(const char *name);

	TriObject *NODE2OBJ(INode *node, int &deleteIt);

	Interface *iface;

	std::vector<INode*> bones;
	std::vector<INode*> nodes;
	std::vector<Modifier*> modifiers;
};

//---------------------------------------------------------------------------
//MAIN EXPORT
//---------------------------------------------------------------------------
void AstEnumProc::exportAll(const char *name) {
	FILE *fmdl = fopen(name, "wt");
	fprintf(fmdl, "#AST_ENGINE_SKINNED_MODEL\n\n");

	//################################################
	//BONES!!
	//################################################

	//-----------------------------------------------
	//Get model bones
	//-----------------------------------------------
	for(int i = 0; i < (int)modifiers.size(); i++) {
		ISkin *skin = (ISkin*)modifiers[i]->GetInterface(I_SKIN);
	
		for(int i = 0; i < skin->GetNumBones(); i++) {
			INode *bone = skin->GetBone(i);
			int j;
			for(j = 0; j < (int)bones.size(); j++) {
				if(bones[j] == bone) break;	
			};

			if(j != bones.size()) continue;
			bones.push_back(bone);
		};
	};
		
	//-------------------------------------------------
	//write bones
	//-----------------------------------------------
	fprintf(fmdl, "num_bones %i\n", bones.size());

	for(int i = 0; i < (int)bones.size(); i++) {
		INode *parent = bones[i]->GetParentNode();
		int j;
		for(j = 0; j < (int)bones.size(); j++) {
			if(bones[j] == parent) break;
		};
		String boneName = bones[i]->GetName();
		replaceSpaces(boneName);

		if(j == bones.size()) {
			fprintf(fmdl, "%s -1\n", boneName.c_str());
		} else {
			fprintf(fmdl, "%s %i\n", boneName.c_str(), j);
		};
	};
	fprintf(fmdl, "\n");
	

	//################################################
	//SUBSETS!!
	//################################################
	
	//-------------------------------------------------
	//get number of subsets
	//-----------------------------------------------
	int nobj = 0;
	for(int n = 0; n < nodes.size(); n++) {
		INode *node = nodes[n];
		int d; TriObject *object = NODE2OBJ(node, d); 
		if (!object) continue;
		nobj++;
	};
	
	//-------------------------------------------------
	//write number of subsets
	//-----------------------------------------------
	fprintf(fmdl, "num_subsets %i\n", nobj);

	//-------------------------------------------------
	//open material list file for writing
	//-----------------------------------------------
	String ml = String((char*)name);
	cutFileExt(ml);
	ml = ml + ".amtrlst";
	FILE *fmtrlst = fopen(ml.c_str(), "wt");

	//-----------------------------------------------
	//get model position
	//-----------------------------------------------
	Point3 min, max;
	for(int n = 0; n < nobj; n++) {
		INode *node = nodes[n];
		int d; TriObject *object = NODE2OBJ(node, d); 
		if (!object) continue;

		object->mesh.buildBoundingBox();
		Box3 box = object->mesh.getBoundingBox();

		if(n == 0) {
			min.x = box.Min().x;
			min.y = box.Min().y;
			min.z = box.Min().z;

			max.x = box.Max().x;
			max.y = box.Max().y;
			max.z = box.Max().z;
		} else {
			if(min.x > box.Min().x) min.x = box.Min().x;
			if(min.y > box.Min().y) min.y = box.Min().y;
			if(min.z > box.Min().z) min.z = box.Min().z;
			
			if(max.x < box.Max().x) max.x = box.Max().x;
			if(max.y < box.Max().y) max.y = box.Max().y;
			if(max.z < box.Max().z) max.z = box.Max().z;
		};
	};
	Point3 pos;
	pos.x = (min.x + max.x) * 0.5;
	pos.y = (min.y + max.y) * 0.5;
	pos.z = (min.z + max.z) * 0.5;

	//-------------------------------------------------
	//process subsets
	//-----------------------------------------------
	for(int n = 0; n < nodes.size(); n++) {
		INode *node = nodes[n];
		int d; TriObject *object = NODE2OBJ(node, d); 
		if (!object) continue;

		ISkin *skin = (ISkin*)modifiers[n]->GetInterface(I_SKIN);
		ISkinContextData *skinData = skin->GetContextInterface(nodes[n]);

		String subsetName = strlwr((char*)node->GetName());
		replaceSpaces(subsetName);
		
		//-------------------------------------------------
		//read model data
		//-----------------------------------------------
		Matrix3 tm = node->GetObjTMAfterWSM(iface->GetTime());
		Matrix3 nm = tm;
		nm.NoScale();
		nm.NoTrans();

		object->mesh.buildNormals();
			
		Vertex *m_vertices = new Vertex[object->mesh.numFaces*3];
		
		for(int i = 0; i < object->mesh.numFaces; i++) {
			Face *f = &object->mesh.faces[i];
			TVFace *tf = &object->mesh.tvFace[i];

			for(int j = 0; j < 3; j++) {
				Point3 v = tm * object->mesh.verts[f->v[j]];
				m_vertices[3*i + j].position.x = v.x - pos.x;
				m_vertices[3*i + j].position.y = v.y - pos.y;
				m_vertices[3*i + j].position.z = v.z - pos.z;

				m_vertices[3*i + j].texcoord.x = object->mesh.tVerts[tf->t[j]].x;
				m_vertices[3*i + j].texcoord.y = object->mesh.tVerts[tf->t[j]].y;


				RVertex *rv = object->mesh.getRVertPtr(f->v[j]);
				int nnormals;
				
				if(rv->rFlags & SPECIFIED_NORMAL) {
					 Point3 normal = nm * rv->rn.getNormal();
					 m_vertices[3*i + j].normal.x = normal.x;
					 m_vertices[3*i + j].normal.y = normal.y;
					 m_vertices[3*i + j].normal.z = normal.z;
				} else {
					if((nnormals = rv->rFlags & NORCT_MASK) && f->smGroup) {
						if(nnormals == 1) {
							Point3 normal = nm * rv->rn.getNormal();
							m_vertices[3*i + j].normal.x = normal.x;
							m_vertices[3*i + j].normal.y = normal.y;
							m_vertices[3*i + j].normal.z = normal.z;
						} else {
							for(int l = 0; l < nnormals; l++) {
								if(rv->ern[l].getSmGroup() & f->smGroup) {
									Point3 normal = nm * rv->ern[l].getNormal();
									m_vertices[3*i + j].normal.x = normal.x;
									m_vertices[3*i + j].normal.y = normal.y;
									m_vertices[3*i + j].normal.z = normal.z;
								};
							};
						};
					} else {
						//Point3 normal = nm * object->mesh.getFaceNormal(i);
						m_vertices[3*i + j].normal.x = 0;
						m_vertices[3*i + j].normal.y = 0;
						m_vertices[3*i + j].normal.z = 0;
						m_vertices[3*i + j].flag = f->v[j];
					};
				};

			};
		};

		//-------------------------------------------------
		//process normals
		//-----------------------------------------------
		for(int i = 0; i < object->mesh.numFaces; i++) {
			Face *f = &object->mesh.faces[i];
			Point3 normal = nm * object->mesh.getFaceNormal(i);
	
			for(int k = 0; k < object->mesh.numFaces*3; k++) {
				if(m_vertices[k].flag == f->v[0] || m_vertices[k].flag == f->v[1] || m_vertices[k].flag == f->v[2]) {
					m_vertices[k].normal.x += normal.x;
					m_vertices[k].normal.y += normal.y;
					m_vertices[k].normal.z += normal.z;
				};
			};
		};

		for(int k = 0; k < object->mesh.numFaces*3; k++) {
			m_vertices[k].normal = m_vertices[k].normal.Normalize();
		};

		//-------------------------------------------------
		//get model indices
		//-----------------------------------------------
		Vertex *vertices = new Vertex[object->mesh.numFaces*3];
		int *indices = new int[object->mesh.numFaces*3];

		int numFaces = object->mesh.numFaces;
		int numVertices = 0;

		for(int i = 0; i < numFaces*3; i++) {
			bool flag = false;
			int index;
			for(int j = 0; j < numVertices; j++) {
				if((vertices[j].position == m_vertices[i].position) && 
					(vertices[j].texcoord == m_vertices[i].texcoord) && 
					(vertices[j].normal == m_vertices[i].normal)) {
					flag = true;
					index = j;
				};
			};

			if(flag) {
				indices[i] = index;
			} else {
				vertices[numVertices].position = m_vertices[i].position;
				vertices[numVertices].texcoord = m_vertices[i].texcoord;
				vertices[numVertices].normal   = m_vertices[i].normal;
				vertices[numVertices].id = m_vertices[i].id;
				indices[i] = numVertices;
				numVertices++;
			};
		};

		//-------------------------------------------------
		//process weights
		//-----------------------------------------------
		for(int j = 0; j < numVertices; j++) {
			int v = vertices[j].id;
			vertices[j].numWeights = skinData->GetNumAssignedBones(v);

			for(int k = 0; k < vertices[j].numWeights; k++) {
				INode *bone = skin->GetBone(skinData->GetAssignedBone(v, k));
				
				int l;
				for(l = 0; l < (int)bones.size(); l++) {
					if(bones[l] == bone) break;
				};
				vertices[j].weights[k].bone = l;
				vertices[j].weights[k].weight = skinData->GetBoneWeight(v, k);

				Matrix3 transform = bone->GetObjTMAfterWSM(iface->GetTime());
				if(transform.Parity()) {
					transform.Invert();
					Matrix3 m;
					m = transform;
					transform.Zero();
					transform -= m;
				} else {
					transform.Invert();
				};

				Matrix3 rotate = transform;
				rotate.Orthogonalize();
				rotate.NoScale();
				rotate.NoTrans();

				vertices[j].weights[k].position = transform * vertices[j].position;
				vertices[j].weights[k].normal = rotate * vertices[j].normal;
			};
		};
				
		//-------------------------------------------------
		//save subset to file
		//-----------------------------------------------
		fprintf(fmdl, "subset %s\n", subsetName.c_str());
	
		fprintf(fmdl, "num_vertices %i\n", numVertices);
				
		for(int v = 0; v < numVertices; v++)	{
			fprintf(fmdl, "num_weights %i\n", vertices[v].numWeights);

			for(int w = 0; w < vertices[v].numWeights; w++)	{
				fprintf(fmdl, "%i %f %f %f %f %f %f %f\n", 
					vertices[v].weights[w].bone,
					vertices[v].weights[w].weight,
					vertices[v].weights[w].position.x - pos.x,
					vertices[v].weights[w].position.z - pos.z,
					vertices[v].weights[w].position.y - pos.y,
					vertices[v].weights[w].normal.x,
					vertices[v].weights[w].normal.z,
					vertices[v].weights[w].normal.y);
			};
			fprintf(fmdl, "%f %f\n",
					vertices[v].texcoord.x,
					vertices[v].texcoord.y);
		};

		fprintf(fmdl, "num_faces %i\n", numFaces); 
		for(int f = 0; f < numFaces; f++)
		{
			fprintf(fmdl, "%i %i %i\n", 
				indices[3*f+0],
				indices[3*f+2],
				indices[3*f+1]);
		};

		fprintf(fmdl, "\n");

		delete [] m_vertices;
		delete [] vertices;
		delete [] indices;

		//################################################
		//Save materials
		//################################################

		Mtl *m = node->GetMtl();
		if(!m || m->ClassID() != Class_ID(DMTL_CLASS_ID, 0)) {
		} else {
			String tex0;
			Texmap *tmap = m->GetSubTexmap(ID_DI);

			if (!tmap || tmap->ClassID() != Class_ID(BMTEX_CLASS_ID, 0)) { 
				tex0 = "";
			} else {
				BitmapTex *bmt = (BitmapTex *)tmap;
				tex0 = strlwr((char *)bmt->GetMapName()); 
				getFileName(tex0);
				replaceSpaces(tex0);
				
				FILE *fmtr = fopen(String(subsetName + ".amtr").c_str(), "wt");
				
				fprintf(fmtr, "pass Ambient\n");
				fprintf(fmtr, "{\n");
				fprintf(fmtr, "shader = mesh_ambient.ashd\n");
				fprintf(fmtr, "mat4 u_mvp = scene::mvp\n");
				fprintf(fmtr, "vec3 u_light_color = scene::light_color\n");
				fprintf(fmtr, "sampler2D u_base_map = %s\n", tex0.c_str());
				fprintf(fmtr, "}\n\n");

				fprintf(fmtr, "pass LightOmni\n");
				fprintf(fmtr, "{\n");
				fprintf(fmtr, "shader = mesh_omni.ashd\n");
				fprintf(fmtr, "float u_spec_coef = 0.5\n");
				fprintf(fmtr, "float u_shininess = 24\n");
				fprintf(fmtr, "mat4 u_mvp = scene::mvp\n");
				fprintf(fmtr, "mat4 u_world = scene::world\n");
				fprintf(fmtr, "vec3 u_light_color = scene::light_color\n");
				fprintf(fmtr, "vec3 u_light_pos = scene::light_position\n");
				fprintf(fmtr, "vec3 u_view_pos = scene::view_position\n");
				fprintf(fmtr, "float u_inv_radius = scene::light_iradius\n");
				fprintf(fmtr, "sampler2D u_base_map = %s\n", tex0.c_str());
				fprintf(fmtr, "sampler2D u_bump_map = normal_map() %s\n", tex0.c_str());
				fprintf(fmtr, "sampler2D u_shadow_map = scene::shadow_map\n");
				fprintf(fmtr, "}\n\n");

				fprintf(fmtr, "pass LightSpot\n");
				fprintf(fmtr, "{\n");
				fprintf(fmtr, "shader = mesh_spot.ashd\n");
				fprintf(fmtr, "float u_spec_coef = 0.5\n");
				fprintf(fmtr, "float u_shininess = 24\n");
				fprintf(fmtr, "mat4 u_mvp = scene::mvp\n");
				fprintf(fmtr, "mat4 u_world = scene::world\n");
				fprintf(fmtr, "mat4 u_spot_proj_transform = scene::spot_transform\n");
				fprintf(fmtr, "vec3 u_light_color = scene::light_color\n");
				fprintf(fmtr, "vec3 u_light_pos = scene::light_position\n");
				fprintf(fmtr, "vec3 u_light_dir = scene::light_direction\n");
				fprintf(fmtr, "float u_inv_radius = scene::light_iradius\n");
				fprintf(fmtr, "vec3 u_view_pos = scene::view_position\n");
				fprintf(fmtr, "sampler2D u_base_map = %s\n", tex0.c_str());
				fprintf(fmtr, "sampler2D u_bump_map = normal_map() %s\n", tex0.c_str());
				fprintf(fmtr, "sampler2D u_shadow_map = scene::shadow_map\n");
				fprintf(fmtr, "sampler2D u_spot_proj_map = scene::spot_map\n");
				fprintf(fmtr, "}\n\n");

				fprintf(fmtr, "pass LightDirect\n");
				fprintf(fmtr, "{\n");
				fprintf(fmtr, "shader = mesh_direct.ashd\n");
				fprintf(fmtr, "float u_specular = 0.5\n");
				fprintf(fmtr, "float u_shininess = 24\n");
				fprintf(fmtr, "mat4 u_mvp = scene::mvp\n");
				fprintf(fmtr, "mat4 u_world = scene::world\n");
				fprintf(fmtr, "vec3 u_light_color = scene::light_color\n");
				fprintf(fmtr, "vec3 u_light_dir = scene::light_direction\n");
				fprintf(fmtr, "vec3 u_view_pos = scene::view_position\n");
				fprintf(fmtr, "sampler2D u_base_map = %s\n", tex0.c_str());
				fprintf(fmtr, "sampler2D u_bump_map = normal_map() %s\n", tex0.c_str());
				fprintf(fmtr, "}");

				fclose(fmtr);
			};
		};

		//################################################
		//Save material_list
		//################################################
		fprintf(fmtrlst, "%s = %s\n", subsetName.c_str(), String(subsetName + ".amtr").c_str());
	};

	//-------------------------------------------------
	//process and save animation frames to file
	//-----------------------------------------------
	Interval interval = iface->GetAnimRange();
	int startTime = interval.Start();
	int endTime = interval.End();
	int ticksPerFrame = GetTicksPerFrame();
	
	fprintf(fmdl, "num_frames %i\n", (endTime - startTime) / ticksPerFrame);

	for(int time = startTime; time < endTime; time += ticksPerFrame) {
		for(int i = 0; i < (int)bones.size(); i++) {
			INode *bone = bones[i];
			Matrix3 transform = bone->GetObjTMAfterWSM(time);
			Point3 xyz = Point3(0,0,0) * transform;
			transform.Orthogonalize();
			transform.NoScale();
			transform.NoTrans();
			transform.Invert();
			Quat rot(transform);
			fprintf(fmdl, "%f %f %f %f %f %f %f",
				xyz.x - pos.x,
				xyz.z - pos.z,
				xyz.y - pos.y,
				-rot.x,
				-rot.z,
				-rot.y,
				-rot.w);
		};
	};
	fclose(fmdl);
};

//---------------------------------------------------------------------------
//CALLBACK
//---------------------------------------------------------------------------
int AstEnumProc::callback(INode *node) {
	Object *obj = node->GetObjectRef();
	if(!obj) return TREE_CONTINUE;
	if(obj->SuperClassID() != GEN_DERIVOB_CLASS_ID) return TREE_CONTINUE;
	
	IDerivedObject *dobj = static_cast<IDerivedObject*>(obj);
	for(int i = 0; i < dobj->NumModifiers(); i++) {
		Modifier *modifier = dobj->GetModifier(i);
		if(modifier->ClassID() == SKIN_CLASSID) {
			nodes.push_back(node);
			modifiers.push_back(modifier);
			return TREE_CONTINUE;
		};
	};

	return TREE_CONTINUE;
};

//---------------------------------------------------------------------------
//CONVERTS NODE TO TRIOBJ
//---------------------------------------------------------------------------
TriObject *AstEnumProc::NODE2OBJ(INode *node, int &deleteIt) {
	deleteIt = false;
	Object *obj = node->EvalWorldState(iface->GetTime()).obj;
	if (obj->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID,0)))
	{
		TriObject *tri = (TriObject *) obj->ConvertToType(iface->GetTime(), 
			Class_ID(TRIOBJ_CLASS_ID, 0));
		if (obj != tri) deleteIt = true;
		return tri;
	}
	else return NULL;
};








//---------------------------------------------------------------------------
//ASTEXPORTER
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//ASTEXPORTER
//---------------------------------------------------------------------------
class ASTExporter : public SceneExport {
public:
	 int          ExtCount()   { return 1; };
	 const TCHAR* Ext(int i)   { if (i == 0) return _T(".asmdl"); else return _T(""); };
	 const TCHAR* LongDesc()   { return _T("AST_engine skinned model exporter"); };
	 const TCHAR* ShortDesc()  { return _T("AST_skinned_exporter"); };
	 const TCHAR* AuthorName() { return _T("Tsyplyaev Alexander"); };
	 const TCHAR* CopyrightMessage() { return _T("Copyright (C) 2006"); };
	 const TCHAR* OtherMessage1()    { return _T(""); };
	 const TCHAR* OtherMessage2()    { return _T(""); };
	 unsigned int Version()          { return 0.1; };

	 void ShowAbout(HWND hWnd) { MessageBox(hWnd, "About", "AST_engine skinned model exporter", MB_OK); }
	 BOOL SupportsOptions(int ext, DWORD options) { return false; };

	 int DoExport(const TCHAR *name, ExpInterface *ei, Interface *i, BOOL suppressPromts = FALSE, DWORD options = 0);
	 
	 ASTExporter() {};
	 virtual ~ASTExporter() {};
};

//---------------------------------------------------------------------------
//DO EXPORT
//---------------------------------------------------------------------------
int ASTExporter::DoExport(const TCHAR *name, ExpInterface *ei, Interface *i, BOOL suppressPromts, DWORD options) {
	AstEnumProc astProc(i);

	ei->theScene->EnumTree(&astProc);

	astProc.exportAll(name);
	return 1;
};










//---------------------------------------------------------------------------
//plugin entry
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//plugin entry
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//plugin entry
//---------------------------------------------------------------------------
HINSTANCE hInstance;
int controlsInit = FALSE;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, ULONG fdwReason, LPVOID lpvReserved) {
	hInstance = hinstDLL;
	if(!controlsInit) {
		controlsInit = TRUE;
		InitCustomControls(hInstance);
		InitCommonControls();
	};
	return TRUE;
};

//---------------------------------------------------------------------------
//Class description
//---------------------------------------------------------------------------
class ASTClassDesc : public ClassDesc {
public:
	int IsPublic() { return 1; }
	void *Create(BOOL loading = FALSE) { return new ASTExporter; };

	const TCHAR *ClassName() { return "AST_engine model exporter"; };
	SClass_ID SuperClassID() { return SCENE_EXPORT_CLASS_ID; };
	Class_ID ClassID()       { return Class_ID(0x150c16c4, 0x51732a4f); };
	const TCHAR *Category()  { return ""; };
};

//---------------------------------------------------------------------------
//ASTClassDesc
//---------------------------------------------------------------------------
static ASTClassDesc astEngineDesc;

//---------------------------------------------------------------------------
//Number of plugins
//---------------------------------------------------------------------------
__declspec(dllexport) int LibNumberClasses() {
	return 1;
};

//---------------------------------------------------------------------------
//LibClassDesc
//---------------------------------------------------------------------------
__declspec(dllexport) ClassDesc *LibClassDesc(int i) {
	return (i == 0) ? &astEngineDesc : 0;
};

//---------------------------------------------------------------------------
//LibDesc
//---------------------------------------------------------------------------
__declspec(dllexport) const TCHAR *LibDescription() {
	return _T("AST engine export plugin");
};

//---------------------------------------------------------------------------
//LibVersion
//---------------------------------------------------------------------------
__declspec(dllexport) ULONG LibVersion() {
	return VERSION_3DSMAX;
};

//---------------------------------------------------------------------------
//????????????
//---------------------------------------------------------------------------
__declspec(dllexport )ULONG CanAutoDefer() {
	return 1;
};

